import math
from .config import pg
from .Utils.utils import getRoundPrefix, submissions
from .Utils.color import colormsg
from sty import fg
import time
import re


def getSeriesPrefix(name):
    if not name:
        return ""
    name = name.strip()
    indefinite = re.compile(f"\s*{name}\s*#?\s*(\d+)\s*")
    definite = re.compile(f"\s*{name}\s*#?\s*(\d+)\s*\/\s*(\d+)\s*")
    for title in submissions():
        defmatch = re.search(definite, title)
        if defmatch:
            return f"[{name} #{int(defmatch.group(1)) + 1}/{defmatch.group(2)}]"
        indefmatch = re.search(indefinite, title)
        if indefmatch:
            return f"[{name} #{int(indefmatch.group(1)) + 1}]"
    return f"[{name} #1]"


def submitRound(r):
    series_prefix = getSeriesPrefix(r.get("series"))
    title = f'{getRoundPrefix()} {series_prefix} {r["title"]}'
    submission = pg.submit(title=title, url=r["url"].strip())

    message = r.get("message")
    if message is not None:
        time.sleep(15)
        submission.reply(message)

        colormsg(f"Message posted to thread: {message}")

    return submission



from .config import pg, username, hints
from .posterino import submitRound
from .hosterino import checkAnswers, checkHints
from .Utils.utils import waitForApproval, approved, postDelay
from .Utils.color import colormsg
from .Loader.loader import getRound
from sty import fg
from importlib.metadata import version
from update_checker import UpdateChecker
from threading import Thread, Event
import time

v = version("botterino")
checker = UpdateChecker()
result = checker.check("botterino", v)
if result:
    colormsg(result, fg.yellow)
    colormsg('run "pip install --upgrade botterino" to update', fg.yellow)


def checkType(r):
    types = []
    if "tolerance" in r and "answer" in r:
        types.append("coordinates")
    if "tolerances" in r and "answers" in r:
        types.append("multiple coordinates")
    if "text" in r and "similarity" in r:
        types.append("text match")
    if "manual" in r:
        types.append("x wrong guesses with manual correct")
    if not types:
        return "no automatic replies"
    if "manual" not in r:
        types.append("automatic")
    return ",".join(types)


def main(stop=None):
    while True:
        colormsg(f"Waiting for {username} to win a round... 🐌", fg.yellow)
        stopped = waitForApproval(stop)
        if stopped or stop:
            colormsg("Stopped botterino", fg.red)
            return
        colormsg(f"Congrats on a well deserved win {username}! ⭐", fg.blue)
        k, r = getRound()
        while not r:
            colormsg(f"No rounds in round file! checking again in 10s", fg.red)
            time.sleep(10)
            k, r = getRound()
        submission = submitRound(r)
        colormsg(
            f"Your round was posted to https://reddit.com{submission.permalink}",
        )
        colormsg(f'Round \'{r["title"]}\' posted in {postDelay()}s', fg.magenta)
        colormsg(f"Checking Answers: {checkType(r)}...", fg.cyan)

        # Create an event to signal the round status
        round_active_event = Event()
        round_active_event.set()

        CheckAnswers = Thread(target=checkAnswers, args=(r, submission))
        CheckHints = Thread(target=checkHints, args=(k, submission, round_active_event))
        CheckAnswers.start()
        CheckHints.start()

        # Wait for threads to finish
        CheckAnswers.join()
        round_active_event.clear()  # Clear the event to signal that the round is over
        CheckHints.join()

        while approved():
            continue
        after = r.get("after")
        if after:
            submission.reply(after)
            colormsg(f"Posted your message after the round: {after}")



import praw
import os
import sys
from .Utils.color import colormsg
from . import botfiles, correctMessage, incorrectMessage, hints
from sty import fg
import random
import webbrowser
import socket


def receive_connection():
    """
    Wait for and then return a connected socket..
    Opens a TCP connection on port 8080, and waits for a single client.
    """
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(("localhost", 8080))
    server.listen(1)
    client = server.accept()[0]
    server.close()
    return client


def send_message(client, message):
    """
    Send message to client and close the connection.
    """
    client.send("HTTP/1.1 200 OK\r\n\r\n{}".format(message).encode("utf-8"))
    client.close()


if sys.platform == "win32":
    os.system("color")

debug = False
roundfile = botfiles.rounds
archivefile = botfiles.archive
hintfile = botfiles.hintfile
cwd = os.getcwd()
# FIXME: clean this up
try:
    os.chdir(botfiles.botconfig)
    reddit = praw.Reddit("botterino")
    if reddit.user.me():
        colormsg(f"Successfully logged into reddit as {reddit.user.me()}", fg.green)
    else:
        colormsg(
            f"Unable to login with username/password. If your account has 2fa continue in browser. Otherwise check {botfiles.prawconfig}",
            fg.yellow,
        )
        os.chdir(botfiles.botconfig)
        reddit = praw.Reddit("botterino", redirect_uri="http://localhost:8080")
        state = str(random.randint(0, 65000))
        scopes = ["identity", "history", "read", "edit", "submit", "privatemessages"]
        url = reddit.auth.url(scopes, state, "permanent")
        colormsg(
            "A window will be opened in the browser to complete the login process to reddit."
        )
        webbrowser.open(url)

        client = receive_connection()
        data = client.recv(1024).decode("utf-8")
        param_tokens = data.split(" ", 2)[1].split("?", 1)[1].split("&")
        params = {
            key: value for (key, value) in [token.split("=") for token in param_tokens]
        }

        if state != params["state"]:
            send_message(
                client,
                "State mismatch. Expected: {} Received: {}".format(
                    state, params["state"]
                ),
            )
        elif "error" in params:
            send_message(client, params["error"])

        refresh_token = reddit.auth.authorize(params["code"])
        send_message(client, "Refresh token: {}".format(refresh_token))
        colormsg(refresh_token, fg.blue)
        if not reddit.user.me():
            colormsg(
                f"Unable to login to reddit. Please check {botfiles.prawconfig}", fg.red
            )
            exit(1)
except Exception as e:
    colormsg(f"Unable to login to reddit. Please check {botfiles.prawconfig}", fg.red)
    colormsg(f"{e}", fg.red)
finally:
    os.chdir(cwd)

pg = reddit.subreddit("itoxtestingfacility" if debug else "picturegame")

username = str(reddit.user.me())

donotreply = {
    "achievements-bot",
    username.lower(),
    "r-picturegame",
    "imreallycuriousbird",
}

api = "https://api.picturega.me"



from tkinter import *
from tkinter import ttk
from .config import roundfile
from .Loader.loader import append, load
from .botterino import main
from sty import fg
from threading import Thread
from .Utils.color import colormsg

root = Tk()
root.title("Botterino")


# Todo: place this in a sensible place
class Stopper:
    def __init__(self):
        self.stopped = False

    def __bool__(self):
        return self.stopped

    def stop(self):
        self.stopped = True

    def unstop(self):
        self.stopped = False


class Runner:
    def __init__(self):
        self.stopper = Stopper()
        self.T = Thread(target=main, args=(self.stopper,))

    def start(self):
        self.stopper.unstop()
        self.T.start()

    def stop(self):
        self.stopper.stop()
        self.T = Thread(target=main, args=(self.stopper,))


R = Runner()


def append_entry():
    name_input = str(name.get())
    if not name_input:
        error_text.set("Name is missing")
        return
    d = load(roundfile)
    if d and name_input in d:
        error_text.set("Name is not unique")
        return
    title_input = title.get()
    if not title_input:
        error_text.set("Title is missing")
        return
    answer_input = answer.get()
    tolerance_input = ""
    if tolerance.get():
        try:
            tolerance_input = int(tolerance.get())
        except:
            error_text.set("Tolerance must be a number")
            return
    if tolerance_input and not answer_input:
        error_text.set("Answer missing when tolerance is present")
        return
    url_input = url.get()
    if not url_input:
        error_text.set("URL is missing")
        return
    manual_input = bool(manual.get())
    data = {}
    inner = {}
    if title_input:
        inner["title"] = title_input
    if answer_input:
        inner["answer"] = answer_input
    if tolerance_input:
        inner["tolerance"] = tolerance_input
    if url_input:
        inner["url"] = url_input
    if manual_input:
        inner["manual"] = manual_input
    data[name_input] = inner
    append(data, roundfile)
    clear_entries()
    colormsg(f"Added round {name_input} to rounds.yaml", fg.green)


def clear_entries():
    name_entry.delete(0, END)
    title_entry.delete(0, END)
    answer_entry.delete(0, END)
    tolerance_entry.delete(0, END)
    url_entry.delete(0, END)
    error_text.set("")
    manual.set(False)


def start_botterino():
    # dummy function to start botterino
    R.start()


def stop_botterino():
    # dummy function to end botterino
    R.stop()


mainframe = ttk.Frame(root, padding="3 6 3 12")
mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
root.columnconfigure(0, weight=1)
root.rowconfigure(0, weight=1)

ttk.Label(mainframe, text="Name").grid(column=1, row=1, sticky=W)
ttk.Label(mainframe, text="Title").grid(column=1, row=2, sticky=W)
ttk.Label(mainframe, text="Answer").grid(column=1, row=3, sticky=W)
ttk.Label(mainframe, text="Tolerance").grid(column=1, row=4, sticky=W)
ttk.Label(mainframe, text="URL").grid(column=1, row=5, sticky=W)
ttk.Label(mainframe, text="Manual").grid(column=1, row=6, sticky=W)

error_text = StringVar()
error_label = ttk.Label(mainframe, foreground="red", textvariable=error_text)
error_label.grid(column=1, row=7, sticky=(W, E))

name = StringVar()
name_entry = ttk.Entry(mainframe, width=50, textvariable=name)
name_entry.grid(column=2, row=1, sticky=(W, E))

title = StringVar()
title_entry = ttk.Entry(mainframe, width=50, textvariable=title)
title_entry.grid(column=2, row=2, sticky=(W, E))

answer = StringVar()
answer_entry = ttk.Entry(mainframe, width=50, textvariable=answer)
answer_entry.grid(column=2, row=3, sticky=(W, E))

tolerance = StringVar()
tolerance_entry = ttk.Entry(mainframe, width=50, textvariable=tolerance)
tolerance_entry.grid(column=2, row=4, sticky=(W, E))

url = StringVar()
url_entry = ttk.Entry(mainframe, width=50, textvariable=url)
url_entry.grid(column=2, row=5, sticky=(W, E))

manual = BooleanVar(value=False)
manual_entry = ttk.Checkbutton(mainframe, variable=manual, onvalue=True, offvalue=False)
manual_entry.grid(column=2, row=6, sticky=(W, E))

ttk.Button(mainframe, text="Clear", command=clear_entries).grid(
    column=2, row=7, sticky=W
)
ttk.Button(mainframe, text="Submit", command=append_entry).grid(
    column=2, row=7, sticky=E
)
ttk.Button(mainframe, text="Start", command=start_botterino).grid(
    column=2, row=8, sticky=W
)
ttk.Button(mainframe, text="Stop", command=stop_botterino).grid(
    column=2, row=8, sticky=E
)

for child in mainframe.winfo_children():
    child.grid_configure(padx=5, pady=2)

root.mainloop()



from sty import fg
import random

badColors = [0, 16, 17, 18, 22, 52, 88, 90] + list(range(232, 256))


def randomColor():
    color = random.randrange(256)
    while color in badColors:
        color = random.randrange(256)
    return fg(color)


def getColorFromAuthor(author):
    seed = hash(author)
    random.seed(seed)
    color = random.randrange(256)
    while color in badColors:
        color = (color + 1) % 256
    return fg(color)


def colormsg(message, color=None, author=None):
    if not color:
        color = randomColor()
    if author:
        color = getColorFromAuthor(author)
    print(f"{color}{message}{fg.rs}")



import folium
import webbrowser
import tempfile
import os
from sty import fg
from ..Utils.color import colormsg

"""
A class for plotting guesses on a map relative to a "correct" guess
"""


class Map:
    def __init__(self, latitude, longitude, zoomLevel=10):
        # TODO: we might want to center the map about the center of all points instead of the correct answer
        self.map = folium.Map(location=[latitude, longitude], zoom_start=zoomLevel)
        self.filePath = ""

    def addPoint(self, latitude, longitude, user, distance, link, color):
        description = """
        {username}

        {distance}m away

        {link}
        """
        folium.Marker(
            location=[latitude, longitude],
            popup=description.format(username=user, distance=distance, link=link),
            icon=folium.Icon(color=color),
        ).add_to(self.map)

    def saveMap(self):
        tempFile = tempfile.NamedTemporaryFile(suffix=".html", delete=False)
        try:
            self.map.save(tempFile)
        except Exception as e:
            colormsg(f"Could not save map: {e}", fg.red)
            return
        self.filePath = tempFile.name

    def getFilePath(self):
        return self.filePath

    def openMapInBrowser(self):
        if not self.filePath:
            return
        webbrowser.open(f"file://{os.path.realpath(self.filePath)}")



import os
import json
from pathlib import Path
from configparser import ConfigParser


class files:
    def __init__(self):
        self.home = os.path.expanduser("~")
        self.botconfig = os.path.join(self.home, "botterino-config")
        self.configfile = os.path.join(self.botconfig, "config.ini")
        self.roundsdir = os.path.join(self.botconfig, "rounds")
        self.rounds = os.path.join(self.roundsdir, "rounds.yaml")
        self.archive = os.path.join(self.roundsdir, "archive.yaml")
        self.prawconfig = os.path.join(self.botconfig, "praw.ini")
        self.hintfile = os.path.join(self.botconfig, "hints.yaml")


botfiles = files()
dirs = [botfiles.botconfig, botfiles.roundsdir]
files = [botfiles.rounds, botfiles.archive, botfiles.prawconfig, botfiles.hintfile]
for d in dirs:
    Path(d).mkdir(parents=True, exist_ok=True)
for f in files:
    with open(f, "a+"):
        pass

if not os.path.exists(botfiles.configfile):
    parser = ConfigParser()
    parser.add_section("config")
    parser.set("config", "correct_message", "+correct")
    parser.set("config", "incorrect_message", "❌")
    parser.set("config", "hints", "[25,45]")
    with open(botfiles.configfile, "w", encoding="utf-8") as f:
        parser.write(f)

parser = ConfigParser()
parser.read(botfiles.configfile, encoding="utf-8")
correctMessage, incorrectMessage = (
    parser["config"]["correct_message"],
    parser["config"]["incorrect_message"],
)
hints = json.loads(parser["config"].get("hints", "[]"))



from .hosterino import checkAnswers, checkHints, checkAnswer
from .config import pg, correctMessage, incorrectMessage, username
from .Utils.color import colormsg
from .Utils.utils import approved, getCurrentComments, hasHostReplied
from .Loader import loader
from sty import fg
import time
from threading import Thread, Event
from datetime import datetime


def processUnrepliedComments(submission, r):
    (
        tolerance,
        manual,
        text,
        answer,
        tolerances,
        answers,
        similarity,
        ignorecase,
    ) = (
        r.get("tolerance"),
        r.get("manual"),
        r.get("text"),
        r.get("answer"),
        r.get("tolerances"),
        r.get("answers"),
        r.get("similarity"),
        r.get("ignorecase"),
    )
    comments = getCurrentComments(submission)
    comments.sort(key=lambda c: datetime.fromtimestamp(c.created_utc))
    for c in comments:
        if (
            not hasHostReplied(c)
            and c.author
            and c.author.name.lower() not in ["r-picturegame", username.lower()]
            and c.is_root
        ):
            if checkAnswer(
                c,
                tolerance,
                text,
                answer,
                tolerances,
                answers,
                similarity,
                ignorecase,
                None,
            ):
                colormsg(f"Correct guess found in comment: {c.permalink}", fg.green)
                if manual:
                    colormsg(
                        f"Guess '{c.body}' looks correct, but you will have to check it out.",
                    )
                else:
                    plusCorrect = c.reply(correctMessage)
                    guesser = c.author.name
                    colormsg(
                        f"Corrected {guesser} in {plusCorrect.created_utc - c.created_utc}s",
                        fg.green,
                    )
                    break
            else:
                colormsg(f"Incorrect guess in comment: {c.permalink}", fg.red)
                c.reply(incorrectMessage)


round = loader.getRound()
while not round:
    colormsg(f"No rounds in round file! checking again in 10s", fg.red)
    time.sleep(10)
    round = loader.getRound()
k, r = round

submission = next(iter(pg.new()))
colormsg(
    f"Checking answers on https://reddit.com{submission.permalink}",
)

# Process unreplied comments
processUnrepliedComments(submission, r)

# Create an event to signal the round status
round_active_event = Event()
round_active_event.set()

CheckAnswers = Thread(target=checkAnswers, args=(r, submission))
CheckHints = Thread(target=checkHints, args=(k, submission, round_active_event))
CheckAnswers.start()
CheckHints.start()

# Wait for threads to finish
CheckAnswers.join()
round_active_event.clear()  # Clear the event to signal that the round is over
CheckHints.join()

while approved():
    continue
after = r.get("after")
if after:
    submission.reply(after)
    colormsg(f"Posted your message after the round: {after}")



from geopy.distance import distance
from geopy.point import Point
from .config import (
    correctMessage,
    incorrectMessage,
    botfiles,
)
import re
from sty import fg
from .Utils.color import colormsg, getColorFromAuthor
from .Utils.utils import (
    decimal,
    getComments,
    getDistance,
    MAPS_URL,
    hyperlink,
    readExistingHints,
)
from difflib import SequenceMatcher
from .Map.map import Map
import time
from threading import Event
from .Loader.loader import loadHints


def withinTolerance(guess, answer, tolerance):
    return distance(guess, answer).m <= tolerance


def checkCoordinateMatch(points, answers, tolerances, used_points=None, depth=0):
    if used_points is None:
        used_points = [False] * len(points)

    if depth == len(points):
        return True

    for i in range(len(points)):
        if not used_points[i] and withinTolerance(points[i], answers[depth], tolerances[depth]):
            used_points[i] = True
            if checkCoordinateMatch(points, answers, tolerances, used_points, depth + 1):
                return True
            used_points[i] = False

    return False


def checkMultipleCoordinates(guess, answers, tolerances):
    guesser = guess.author.name
    answers = [Point(a) for a in answers]
    match = re.findall(decimal, guess.body)

    if len(match) != len(answers):
        colormsg(f"{guesser}'s guess {guess.body} was incorrect", author=guesser)
        return False

    points = [Point(f"{lat},{lon}") for lat, lon in match]

    if checkCoordinateMatch(points, answers, tolerances):
        return True
    else:
        colormsg(f"{guesser}'s guess {guess.body} was incorrect", author=guesser)
        return False


def checkCoordinates(guess, answer, tolerance, map):
    guesser = guess.author.name
    answer = Point(answer)
    errorAndPoint = getDistance(guess.body, answer)
    error, point = errorAndPoint if errorAndPoint else (None, None)
    if error is None:
        colormsg(
            f"Could not find a coordinate in guess '{guess.body}' by {guesser}",
            author=guesser,
        )
        return "ignore"
    error = round(error, 2)
    mapslink = MAPS_URL.format(point.latitude, point.longitude)
    color = fg.green if error <= tolerance else getColorFromAuthor(guesser)
    pl = ""
    if hasattr(guess, "context"):
        pl = guess.context
    elif hasattr(guess, "permalink"):
        pl = guess.permalink
    commentLink = f"https://reddit.com{pl}"
    if error < 1000:
        colormsg(
            f'{guesser}\'s {hyperlink("guess", commentLink)} was {error}m {hyperlink("off", mapslink)}',
            color,
        )
    else:
        colormsg(
            f'{guesser}\'s {hyperlink("guess", commentLink)} was {round(error/1000,2)}km {hyperlink("off", mapslink)}',
            color,
        )

    if map:
        # TODO: try to give different users differnt colors
        map.addPoint(
            point.latitude,
            point.longitude,
            guesser,
            error,
            commentLink,
            "red" if error > tolerance else "green",
        )

    return error <= tolerance


def checkText(guess, answer, tolerance, ignorecase):
    guesser = guess.author.name
    text = guess.body.strip().replace("\\", "")

    if ignorecase:
        text, answer = text.lower(), answer.lower()

    similarity = SequenceMatcher(None, text, answer).ratio()
    colormsg(
        f"{guesser}'s guess was {round(similarity * 100, 3)}% similar to the correct answer",
        author=guesser,
    )
    return similarity >= tolerance


def postHint(submission, time, hintText):
    if not hintText:
        colormsg(f"Skipping {time}m hint: no text provided", fg.yellow)
        return
    hint = submission.reply(f"Hint({time}m): {hintText}")
    colormsg(f"Posted hint ({time}m) to https://reddit.com{hint.permalink}", fg.green)


def checkHints(key, submission, round_active_event, poll_interval=30):
    initial = True
    while round_active_event.is_set():
        hints = loadHints(key)
        existing_hints = readExistingHints(submission)

        for hint in hints:
            duration = int(time.time() - submission.created_utc) // 60
            if hint["time"] <= duration:
                if any(hint["text"] in existing_hint for existing_hint in existing_hints):
                    if initial:
                        colormsg(f"Looks like the hint for time {hint['time']}m is already posted")
                else:
                    postHint(submission, duration, hint["text"])

        initial = False
        time.sleep(poll_interval)


def checkAnswer(
    comment,
    tolerance,
    text,
    answer,
    tolerances,
    answers,
    similarity,
    ignorecase,
    answerPlot,
):
    result = True
    if tolerance is not None:
        tolerance = float(tolerance)
        r = checkCoordinates(comment, answer, tolerance, answerPlot)
        if r == "ignore":
            return False
        result = result and r
    elif tolerances:
        tolerances = [float(t) for t in r["tolerances"]]
        if len(answers) != len(tolerances):
            colormsg(f"Refusing to check answers, number of tolerances must equal number of answers.", fg.red)
        result = result and checkMultipleCoordinates(comment, answers, tolerances)

    if text and similarity is None:
        return False

    if ignorecase is None:
        ignorecase = True

    if text:
        result = result and checkText(comment, text, similarity, ignorecase)

    return result


def checkAnswers(r, submission):
    (
        tolerance,
        manual,
        text,
        answer,
        tolerances,
        answers,
        similarity,
        ignorecase,
    ) = (
        r.get("tolerance"),
        r.get("manual"),
        r.get("text"),
        r.get("answer"),
        r.get("tolerances"),
        r.get("answers"),
        r.get("similarity"),
        r.get("ignorecase"),
    )

    answerPlot = None
    if answer and tolerance and not answers:
        try:
            answerPoint = Point(answer)
            answerPlot = Map(answerPoint.latitude, answerPoint.longitude)
        except Exception:
            pass

    if tolerance is None and tolerances is None and text is None:
        return

    for c in getComments(submission):
        result = checkAnswer(
            c,
            tolerance,
            text,
            answer,
            tolerances,
            answers,
            similarity,
            ignorecase,
            answerPlot,
        )

        if not result:
            c.reply(incorrectMessage)
        if result:
            if manual:
                colormsg(
                    f"Guess '{c.body}' looks correct, but you will have to check it out.",
                )
            else:
                plusCorrect = c.reply(correctMessage)
                guesser = c.author.name
                colormsg(
                    f"Corrected {guesser} in {plusCorrect.created_utc - c.created_utc}s",
                    fg.green,
                )
                break

    # TODO: show the map on every guess instead of only when the round is over
    if answerPlot:
        answerPlot.saveMap()
        outputFile = answerPlot.getFilePath()
        if outputFile:
            colormsg(f"Answers to your round plotted at {outputFile}", fg.cyan)
            answerPlot.openMapInBrowser()
